import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from typing import List, Set

from yanga.project.config import YangaUserConfig


def find_files(start_dir: Path, search_pattern: str, exclude_dirs: List[str] = []) -> Set[Path]:
    """Find files matching the search_pattern within start_dir, excluding the directories in exclude_dirs.

    Args:
        start_dir (Path): The directory to start the search from.
        search_pattern (str): The pattern of file names to search for.
        exclude_dirs (List[str]): A list of relative paths to exclude from the search.

    Returns:
        Set[Path]: A set of Paths to the files found.
    """
    start_dir = start_dir.resolve()
    exclude_paths = {start_dir.joinpath(d) for d in exclude_dirs}
    found_files = set()

    for dirpath, dirnames, filenames in os.walk(start_dir):
        # Modify dirnames in-place when you want to prevent os.walk from descending
        # into certain directories.
        # It's a feature of os.walk that allows for more efficient file system operations.
        dirnames[:] = [d for d in dirnames if not _is_excluded(Path(dirpath).joinpath(d), exclude_paths)]
        for filename in filenames:
            if Path(filename).match(search_pattern):
                found_files.add(Path(dirpath).joinpath(filename))

    return found_files


def _is_excluded(dir_path: Path, exclude_paths: Set[Path]) -> bool:
    """Check if the directory is in the exclude paths set."""
    return any(dir_path == exclude_path or dir_path.is_relative_to(exclude_path) for exclude_path in exclude_paths)


class YangaConfigSlurper:
    """Read all 'yanga.yaml' configuration files from the project."""

    CONFIG_FILE = "yanga.yaml"

    def __init__(self, project_dir: Path, exclude_dirs: List[str] = []) -> None:
        self.project_dir = project_dir
        self.exclude_dirs = exclude_dirs

    def slurp(self) -> List[YangaUserConfig]:
        user_configs = []
        config_files = find_files(self.project_dir, self.CONFIG_FILE, self.exclude_dirs)
        with ThreadPoolExecutor() as executor:
            future_to_file = {
                executor.submit(self.parse_config_file, config_file): config_file for config_file in config_files
            }
            for future in as_completed(future_to_file):
                user_configs.append(future.result())
        return user_configs

    def parse_config_file(self, config_file: Path) -> YangaUserConfig:
        return YangaUserConfig.from_file(config_file)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List

from yanga.ybuild.components import BuildComponent

from .generated_file import GeneratedFile


class Builder(ABC):
    def __init__(self, output_path: Path) -> None:
        self.output_path = output_path

    @abstractmethod
    def with_project_name(self, name: str) -> "Builder":
        ...

    @abstractmethod
    def with_components(self, components: List[BuildComponent]) -> "Builder":
        ...

    @abstractmethod
    def with_include_directories(self, include_directories: List[Path]) -> "Builder":
        ...

    @abstractmethod
    def build(self) -> GeneratedFile:
        ...



import json
import shutil
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional, Union

from cookiecutter.main import cookiecutter
from jinja2 import Environment, FileSystemLoader
from mashumaro import DataClassDictMixin
from py_app_dev.core.cmd_line import Command, register_arguments_for_config_dataclass
from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger, time_it


@dataclass
class TemplateFileConfig:
    # Relative path to input directory
    src: str
    # Relative path to the output directory. None means root directory.
    dest: Optional[str] = None


class ProjectBuilder:
    def __init__(self, project_dir: Path, input_dir: Optional[Path] = None) -> None:
        self.project_dir = project_dir
        self.input_dir = input_dir if input_dir else Path(__file__).parent.joinpath("project_templates")

        self.dirs: List[Path] = []
        self.cookiecutter_dir: Optional[Path] = None
        # Store tuples of (template_path, destination_path)
        # where the destination path is relative to the output directory
        self.template_files: List[TemplateFileConfig] = []
        self.template_config: Dict[str, Any] = {}
        self.check_target_directory_flag = True

    def with_disable_target_directory_check(self) -> "ProjectBuilder":
        self.check_target_directory_flag = False
        return self

    def with_dir(self, dir: Union[Path, str]) -> "ProjectBuilder":
        self.dirs.append(self.resolve_file_path(dir))
        return self

    def with_cookiecutter_dir(self, cookiecutter_dir: Union[Path, str]) -> "ProjectBuilder":
        self.cookiecutter_dir = self.resolve_file_path(cookiecutter_dir)
        return self

    def with_jinja_template(self, template_path: str, dest_path: Optional[str] = None) -> "ProjectBuilder":
        self.template_files.append(TemplateFileConfig(template_path, dest_path))
        return self

    def with_template_config(self, config: Dict[str, Any]) -> "ProjectBuilder":
        self.template_config.update(config)
        return self

    def with_template_config_file(self, json_file: Union[Path, str]) -> "ProjectBuilder":
        json_file_path = self.resolve_file_path(json_file)
        with json_file_path.open() as f:
            self.template_config.update(json.load(f))
        return self

    def resolve_file_paths(self, files: List[Path | str]) -> List[Path]:
        return [self.resolve_file_path(file) for file in files]

    def resolve_file_path(self, file: Union[Path, str]) -> Path:
        return self.input_dir.joinpath(file) if isinstance(file, str) else file

    @staticmethod
    def _create_project_from_template(input_dir: Path, output_dir: Path) -> None:
        with TemporaryDirectory() as tmp_dir:
            tmp_dir_path = Path(tmp_dir)
            project_dir_name = output_dir.name
            cookiecutter(
                input_dir.as_posix(),
                no_input=True,
                output_dir=tmp_dir_path.as_posix(),
                extra_context={"project_dir_name": project_dir_name},
            )
            # Copy the temporary directory to the output directory
            shutil.copytree(tmp_dir_path / project_dir_name, output_dir, dirs_exist_ok=True)

    @staticmethod
    def _check_target_directory(project_dir: Path) -> None:
        if project_dir.is_dir() and any(project_dir.iterdir()):
            raise UserNotificationException(
                f"Project directory '{project_dir}' is not empty."
                " The target directory shall either be empty or not exist."
            )

    def _render_templates(self) -> None:
        env = Environment(loader=FileSystemLoader(self.input_dir), keep_trailing_newline=True)  # nosec
        for template_file in self.template_files:
            template = env.get_template(template_file.src)
            rendered_content = template.render(self.template_config)
            dest_file_path = template_file.dest if template_file.dest else Path(template_file.src).name
            dest_path = self.project_dir / dest_file_path
            dest_path.parent.mkdir(parents=True, exist_ok=True)
            dest_path.write_text(rendered_content)

    def build(self) -> None:
        if self.check_target_directory_flag:
            self._check_target_directory(self.project_dir)
        if self.cookiecutter_dir:
            self._create_project_from_template(self.cookiecutter_dir, self.project_dir)
        for dir in self.dirs:
            shutil.copytree(dir, self.project_dir, dirs_exist_ok=True)
        self._render_templates()  # Render and write Jinja2 templates


@dataclass
class InitCommandConfig(DataClassDictMixin):
    project_dir: Path = field(
        default=Path(".").absolute(),
        metadata={"help": "Project root directory. " "Defaults to the current directory if not specified."},
    )

    bootstrap: Optional[bool] = field(
        default=False,
        metadata={
            "help": "Initialize only the bootstrap files.",
            "action": "store_true",
        },
    )

    @classmethod
    def from_namespace(cls, namespace: Namespace) -> "InitCommandConfig":
        return cls.from_dict(vars(namespace))


class YangaInit:
    def __init__(self, config: InitCommandConfig) -> None:
        self.logger = logger.bind()
        self.config = config

    def run(self) -> None:
        self.logger.info(f"Run yanga init in '{self.config.project_dir.absolute().as_posix()}'")
        project_builder = ProjectBuilder(self.config.project_dir)
        project_builder.with_jinja_template("template/bootstrap_j2.ps1", "bootstrap.ps1").with_jinja_template(
            "template/bootstrap_j2.py", "bootstrap.py"
        ).with_jinja_template("template/bootstrap_j2.json", "bootstrap.json").with_template_config_file(
            "template/cookiecutter.json"
        )
        if self.config.bootstrap:
            project_builder.with_disable_target_directory_check()
        else:
            project_builder.with_dir("common").with_cookiecutter_dir("template")
            project_builder.with_dir("mini")
        project_builder.build()


class InitCommand(Command):
    def __init__(self) -> None:
        super().__init__("init", "Init a yanga project")
        self.logger = logger.bind()

    @time_it("Init")
    def run(self, args: Namespace) -> int:
        self.logger.info(f"Running {self.name} with args {args}")
        YangaInit(InitCommandConfig.from_namespace(args)).run()
        return 0

    def _register_arguments(self, parser: ArgumentParser) -> None:
        register_arguments_for_config_dataclass(parser, InitCommandConfig)



from dataclasses import dataclass
from enum import Enum
from typing import Optional


class BuildSystemCommand(Enum):
    NONE = "none"
    ALL = "all"
    CUSTOM = "custom"
    CLEAN = "clean"
    BUILD = "build"
    TEST = "test"
    COMPILE = "compile"


@dataclass
class BuildSystemRequest:
    variant_name: str
    component_name: Optional[str] = None
    command: BuildSystemCommand = BuildSystemCommand.ALL

    @property
    def target_name(self) -> str:
        if self.component_name:
            return f"{self.component_name}_{self.command.value}"
        return self.command.value if self.command else "all"


class CustomBuildSystemRequest(BuildSystemRequest):
    def __init__(self, variant_name: str, custom_request: str) -> None:
        super().__init__(variant_name, None, BuildSystemCommand.CUSTOM)
        self.custom_request = custom_request

    @property
    def target_name(self) -> str:
        return self.custom_request


class BuildVariantRequest(BuildSystemRequest):
    def __init__(self, variant_name: str) -> None:
        super().__init__(variant_name, None, BuildSystemCommand.BUILD)


class TestVariantRequest(BuildSystemRequest):
    def __init__(self, variant_name: str) -> None:
        super().__init__(variant_name, None, BuildSystemCommand.TEST)


class CleanVariantRequest(BuildSystemRequest):
    def __init__(self, variant_name: str) -> None:
        super().__init__(variant_name, None, BuildSystemCommand.CLEAN)


class CompileComponentRequest(BuildSystemRequest):
    def __init__(self, variant_name: str, component_name: str) -> None:
        super().__init__(variant_name, component_name, BuildSystemCommand.COMPILE)


class TestComponentRequest(BuildSystemRequest):
    def __init__(self, variant_name: str, component_name: str) -> None:
        super().__init__(variant_name, component_name, BuildSystemCommand.TEST)



import sys
from pathlib import Path


class ProjectArtifactsLocator:
    """Provides paths to project artifacts."""

    config_file_name = "yanga.yaml"

    def __init__(self, project_root_dir: Path) -> None:
        self.project_root_dir = project_root_dir
        self.build_dir = project_root_dir / "build"
        self.variants_dir = project_root_dir / "variants"
        self.platforms_dir = project_root_dir / "platforms"
        self.config_file = project_root_dir / self.config_file_name


class ProjectBuildArtifactsLocator(ProjectArtifactsLocator):
    def __init__(
        self,
        project_root_dir: Path,
        variant_name: str,
    ) -> None:
        super().__init__(project_root_dir)
        self.variant_build_dir = self.build_dir / variant_name
        self.variant_dir = self.variants_dir / variant_name
        self.external_dependencies_dir = self.build_dir / "external"
        scripts_dir = "Scripts" if sys.platform.startswith("win32") else "bin"
        self.venv_scripts_dir = self.project_root_dir.joinpath(".venv").joinpath(scripts_dir)



from enum import Enum
from pathlib import Path

import customtkinter
from PIL import Image


class Icons(Enum):
    YANGA_ICON = "yanga.ico"

    def __init__(self, file_name: str) -> None:
        self.resources_dir = Path(__file__).parent.joinpath("resources")
        self.file = self.resources_dir.joinpath(file_name).as_posix()

    @property
    def image(self) -> customtkinter.CTkImage:
        return customtkinter.CTkImage(
            dark_image=Image.open(self.file),
            size=(30, 30),
        )



import importlib
import shutil
from abc import ABC
from dataclasses import dataclass
from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from pathlib import Path
from typing import List, Type

from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger
from py_app_dev.core.runnable import Executor, Runnable

from yanga.project.config import PipelineConfig, StageConfig

from .environment import BuildEnvironment


class Stage(Runnable, ABC):
    def __init__(self, environment: BuildEnvironment, group_name: str) -> None:
        self.environment = environment
        # TODO: I do not like that a Stage knows about a variant and that it has a build dir
        self.output_dir = self.environment.artifacts_locator.variant_build_dir / group_name

    @property
    def project_root_dir(self) -> Path:
        return self.environment.project_root_dir


@dataclass
class BuildStage:
    """TODO: what is the difference between this class and Stage? The name is very strange"""

    group_name: str
    _class: Type[Stage]


class PipelineLoader:
    def __init__(self, pipeline_config: PipelineConfig, project_root_dir: Path) -> None:
        self.pipeline_config = pipeline_config
        self.project_root_dir = project_root_dir

    def load_stages(self) -> List[BuildStage]:
        result = []
        for group_name, stages_config in self.pipeline_config.items():
            result.extend(self._load_stages(group_name, stages_config, self.project_root_dir))
        return result

    @staticmethod
    def _load_stages(group_name: str, stages_config: List[StageConfig], project_root_dir: Path) -> List[BuildStage]:
        result = []
        for stage_config in stages_config:
            stage_class_name = stage_config.class_name or stage_config.stage
            if stage_config.module:
                stage_class = PipelineLoader._load_module_stage(stage_config.module, stage_class_name)
            elif not stage_config.file:
                # no file means that the stage is a built-in stage
                stage_class = PipelineLoader._load_builtin_stage(stage_class_name)
            else:
                stage_class = PipelineLoader._load_user_stage(
                    project_root_dir.joinpath(stage_config.file), stage_class_name
                )
            result.append(BuildStage(group_name, stage_class))
        return result

    @staticmethod
    def _load_builtin_stage(stage_class_name: str) -> Type[Stage]:
        stage_module = import_module("yanga.ybuild.stages")
        try:
            stage_class = getattr(stage_module, stage_class_name)
        except AttributeError:
            raise UserNotificationException(
                f"Stage '{stage_class_name}' is not a Yanga built-in stage."
                " Please check your pipeline configuration."
            )
        return stage_class

    @staticmethod
    def _load_user_stage(python_file: Path, stage_class_name: str) -> Type[Stage]:
        # Create a module specification from the file path
        spec = spec_from_file_location(f"user__{stage_class_name}", python_file)
        if spec and spec.loader:
            stage_module = module_from_spec(spec)
            # Import the module
            spec.loader.exec_module(stage_module)
            try:
                stage_class = getattr(stage_module, stage_class_name)
            except AttributeError:
                raise UserNotificationException(
                    f"Could not load class '{stage_class_name}' from file '{python_file}'."
                    " Please check your pipeline configuration."
                )
            return stage_class
        raise UserNotificationException(
            f"Could not load file '{python_file}'." " Please check the file for any errors."
        )

    @staticmethod
    def _load_module_stage(module_name: str, stage_class_name: str) -> Type[Stage]:
        try:
            module = importlib.import_module(module_name)
            stage_class = getattr(module, stage_class_name)
        except ImportError:
            raise UserNotificationException(
                f"Could not load module '{module_name}'. Please check your pipeline configuration."
            )
        except AttributeError:
            raise UserNotificationException(
                f"Could not load class '{stage_class_name}' from module '{module_name}'."
                " Please check your pipeline configuration."
            )
        return stage_class


class StageRunner:
    """It checks if a stage must run in current environment.
    A stage shall run if any of the dependencies changed or one of the outputs is missing.
    All dependencies and outputs relevant information is stored in the stage output directory
    in a <stage>.deps file.
    """

    def __init__(self, environment: BuildEnvironment, stage: BuildStage) -> None:
        self.logger = logger.bind()
        self.environment = environment
        self.stage = stage

    def run(self) -> None:
        stage = self.stage._class(self.environment, self.stage.group_name)
        if self.environment.is_clean_required():
            self.logger.info(f"Cleaning stage '{stage.__class__.__name__}'")
            shutil.rmtree(stage.output_dir, ignore_errors=True)
        else:
            stage.output_dir.mkdir(parents=True, exist_ok=True)
            Executor(stage.output_dir).execute(stage)



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, TypeAlias

from py_app_dev.core.exceptions import UserNotificationException
from py_app_dev.core.logging import logger

from yanga.ybuild.components import BuildComponent, BuildComponentType
from yanga.ybuild.pipeline import BuildStage, PipelineLoader

from .config import ComponentConfig, PipelineConfig, VariantConfig, YangaUserConfig
from .config_slurper import YangaConfigSlurper


@dataclass
class ComponentConfigWithLocation:
    config: ComponentConfig
    #: Path to the file where the component is defined
    file: Optional[Path] = None


ComponentsConfigsPool: TypeAlias = Dict[str, ComponentConfigWithLocation]


class YangaProjectSlurper:
    def __init__(self, project_dir: Path) -> None:
        self.logger = logger.bind()
        self.project_dir = project_dir
        self.user_configs: List[YangaUserConfig] = YangaConfigSlurper(
            self.project_dir, [".git", ".github", ".vscode", "build", ".venv"]
        ).slurp()
        self.components_configs_pool: ComponentsConfigsPool = self._collect_components_configs(self.user_configs)
        self.stages: List[BuildStage] = self._collect_stages(self.user_configs, self.project_dir)
        self.variants: List[VariantConfig] = self._collect_variants(self.user_configs)
        self.print_project_info()

    @property
    def user_config_files(self) -> List[Path]:
        return [user_config.file for user_config in self.user_configs if user_config.file]

    def print_project_info(self) -> None:
        self.logger.info("-" * 80)
        self.logger.info(f"Project directory: {self.project_dir}")
        self.logger.info(f"Parsed {len(self.user_configs)} configuration file(s).")
        self.logger.info(f"Found {len(self.components_configs_pool.values())} component(s).")
        self.logger.info(f"Found {len(self.stages)} stage(s).")
        self.logger.info(f"Found {len(self.variants)} variant(s).")
        self.logger.info("-" * 80)

    def get_variant_config(self, variant_name: str) -> VariantConfig:
        variant = next((v for v in self.variants if v.name == variant_name), None)
        if not variant:
            raise UserNotificationException(f"Variant '{variant_name}' not found in the configuration.")

        return variant

    def get_variant_config_file(self, variant_name: str) -> Optional[Path]:
        variant = self.get_variant_config(variant_name)
        return self.project_dir.joinpath(variant.config_file) if variant.config_file else None

    def get_variant_components(self, variant_name: str) -> List[BuildComponent]:
        return self._collect_variant_components(self.get_variant_config(variant_name))

    def _collect_variant_components(self, variant: VariantConfig) -> List[BuildComponent]:
        """ "Collect all components for the given variant.
        Look for components in the component pool and add them to the list."""
        components = []
        if not variant.bom:
            raise UserNotificationException(f"Variant '{variant.name}' is empty (no 'bom' found).")
        for component_name in variant.bom.components:
            component_config = self.components_configs_pool.get(component_name, None)
            if not component_config:
                raise UserNotificationException(f"Component '{component_name}' not found in the configuration.")
            components.append(self._create_build_component(component_config))
        self._resolve_subcomponents(components, self.components_configs_pool)
        return components

    def _create_build_component(self, component_config: ComponentConfigWithLocation) -> BuildComponent:
        # TODO: determine component type based on if it has sources, subcomponents
        component_type = BuildComponentType.COMPONENT
        component_path = component_config.file.parent if component_config.file else self.project_dir
        build_component = BuildComponent(
            component_config.config.name,
            component_type,
            component_path,
        )
        if component_config.config.sources:
            build_component.sources = component_config.config.sources
        if component_config.config.test_sources:
            build_component.test_sources = component_config.config.test_sources
        return build_component

    def _collect_components_configs(self, user_configs: List[YangaUserConfig]) -> ComponentsConfigsPool:
        components_config: ComponentsConfigsPool = {}
        for user_config in user_configs:
            for component_config in user_config.components:
                if components_config.get(component_config.name, None):
                    # TODO: throw the UserNotificationException and mention the two files
                    #  where the components are defined
                    raise UserNotificationException(
                        f"Component '{component_config.name}' is defined in multiple configuration files."
                        f"See {components_config[component_config.name].file} and {user_config.file}"
                    )
                components_config[component_config.name] = ComponentConfigWithLocation(
                    component_config, user_config.file
                )
        return components_config

    def _resolve_subcomponents(
        self,
        components: List[BuildComponent],
        components_configs_pool: ComponentsConfigsPool,
    ) -> None:
        """Resolve subcomponents for each component."""
        components_pool = {c.name: c for c in components}
        for component in components:
            # It can not be that there is no configuration for the component,
            # otherwise it would not be in the list
            component_config = components_configs_pool.get(component.name)
            if component_config and component_config.config.components:
                for subcomponent_name in component_config.config.components:
                    subcomponent = components_pool.get(subcomponent_name, None)
                    if not subcomponent:
                        # TODO: throw the UserNotificationException and mention the file
                        # where the subcomponent was defined
                        raise UserNotificationException(
                            f"Component '{subcomponent_name}' not found in the configuration."
                        )
                    component.components.append(subcomponent)
                    subcomponent.is_subcomponent = True

    def _collect_stages(self, user_configs: List[YangaUserConfig], project_root_dir: Path) -> List[BuildStage]:
        """Find the pipeline configuration and collect all stages.
        In case there are multiple pipeline configurations, throw an exception.
        """
        pipeline_config = self._find_pipeline_config(user_configs)
        return PipelineLoader(pipeline_config, project_root_dir).load_stages()

    def _find_pipeline_config(self, user_configs: List[YangaUserConfig]) -> PipelineConfig:
        configs = [user_config.pipeline for user_config in user_configs if user_config.pipeline]
        if not configs:
            raise UserNotificationException("No pipeline configuration found.")
        elif len(configs) > 1:
            raise UserNotificationException(
                "Multiple pipeline configurations found. " "Only one pipeline configuration is allowed."
            )
        return configs[0]

    def _collect_variants(self, user_configs: List[YangaUserConfig]) -> List[VariantConfig]:
        variants = []
        for user_config in user_configs:
            variants.extend(user_config.variants)
        return variants
